﻿/**
 *	Helper class for lamejs.
 *	You must include lamejs/lame.all.js.
 */
function LameEncoderHelper(onDataReady)
{
	/*
	 *	PRIVATE
	 */
	 
	var mp3Encoder, config;
	var MAX_SAMPLES = 1152;
	var nChannels = 1;
	var nSampleRate = 44100;
	var nBitRate = 128;

	function appendToBuffer(mp3Buf)
	{
		onDataReady(mp3Buf);
	}
	
	function convertBuffer(arrayBuffer)
	{
		var input = new Float32Array(arrayBuffer);
		var output = new Int16Array(arrayBuffer.length);
		for (var i = 0; i < input.length; i++)
		{
			var s = Math.max(-1, Math.min(1, input[i]));
			output[i] = (s < 0 ? s * 0x8000 : s * 0x7FFF);
		}
		return output;
	};
	
	/*
	 *	PUBLIC
	 */
	
	this.init = function (prefConfig)
	{
		config = prefConfig || {debug: true};
		nSampleRate = config.sampleRate || 44100;
		nBitRate = config.bitRate || 128;
		nChannels = config.numberOfChannels || 1;
		mp3Encoder = new lamejs.Mp3Encoder(nChannels, nSampleRate, nBitRate);
		console.log("Init lamejs: sampleRate=" + nSampleRate + ", bitRate=" + nBitRate);
	};
	
	this.encode = function(inputBuffer)
	{
		var samplesLeft = convertBuffer(inputBuffer[0]);
		var samplesRight;
		if (inputBuffer.length > 1)
			samplesRight = convertBuffer(inputBuffer[1]);
		else if (nChannels > 1)
			samplesRight = samplesLeft;
		var remaining = samplesLeft.length;
		for (var i = 0; remaining >= 0; i += MAX_SAMPLES)
		{
			var left = samplesLeft.subarray(i, i + MAX_SAMPLES);
			var right = nChannels > 1 ? samplesRight.subarray(i, i + MAX_SAMPLES) : null;
			var mp3buf = mp3Encoder.encodeBuffer(left, right);
			appendToBuffer(mp3buf);
			remaining -= MAX_SAMPLES;
		}
	};
	
	this.getSampleRate = function() { return nSampleRate; };
	
	this.getBitRate = function() { return nBitRate; };
	
	this.getNumChannels = function() { return nChannels; };
}

/**
 *	Helper class for ogg opus codec.
 *	You must place script 'Recorderjs-0.5.0/build/encoderWorker.min.js' with current file.
 */
function OpusEncoderHelper(onDataReady, encoderPath)
{
	/*
	 *	PRIVATE
	 */

	var sEncoderPath = encoderPath || 'encoderWorker.min.js';
	var objEncoder;
	var objConfig;
	
	var bStarted = false;
	/*
	 *	PUBLIC
	 */
	
	this.init = function (prefConfig)
	{
		objConfig = prefConfig;
		objConfig.command = "init";
		objConfig.bufferLength = prefConfig.bufferLength || 4096;
		objConfig.numberOfChannels = prefConfig.numberOfChannels || 1;
		objConfig.monitorGain = prefConfig.monitorGain || 0;
		objConfig.bitRate = prefConfig.bitRate || 96000;
		objConfig.originalSampleRate = prefConfig.originalSampleRate || 48000;
		objConfig.encoderSampleRate = prefConfig.sampleRate || 48000;
		objConfig.streamPages = true;
		objConfig.leaveStreamOpen = false;
		objConfig.maxBuffersPerPage = prefConfig.maxBuffersPerPage || 1;
		objConfig.encoderApplication = prefConfig.encoderApplication || 2049; /* music*/
		objConfig.encoderFrameSize = prefConfig.encoderFrameSize || 60; /* Opus can encode frames of 2.5, 5, 10, 20, 40, or 60 ms.*/
		objConfig.resampleQuality = prefConfig.resampleQuality || 5; /* 0 - 10*/
		objEncoder = new Worker(sEncoderPath);
		bStarted = false;
		objEncoder.addEventListener("message", function( e ) { bStarted = true; onDataReady( e.data ); });
		objEncoder.postMessage(objConfig);
	};
	
	this.encode = function (aBuffers)
	{
		if (!bStarted) return;
		objEncoder.postMessage({ 
			command: "encode",
			buffers: aBuffers
		});
	};
	
	this.getSampleRate = function() { return objConfig.encoderSampleRate; };
	
	this.getBitRate = function() { return objConfig.bitRate / 1000; };
	
	this.getNumChannels = function() { return objConfig.numberOfChannels; };
}

/**
 *	Helper class for raw data.
 */
function RawEncoderHelper(onDataReady)
{
	/*
	 *	PRIVATE
	 */

    var buffer = null;
    var bufLen = 0;
    var bufPos = 0;
    var origSampleRate = 0;
    var encoderSampleRate = 0;
    var bitRate = 0;
    var sampleFmt ='';
    var nSrc = 0;
    var nRes = 0;
    var nLastRD = 0;    /* resample correction for buffer concatenation*/
    var nLastVal = 0;   /* last value for resample correction*/
    var bIs16Bit = 0;
    /* compression*/
    var compressedBuff = null; /* Uint8Array*/
    var tmpBuff = new Int16Array(4);

    /* stats*/
    var totalCompressed = 0;
    var totalData = 0;

    /**
     *  Audio data compression
     */ 
    function compressBuffer(buffer)
    {
        /*"use asm";*/
        var maxd = 0;
        var num  = 0;
        var pos = 2;    /* marker pos*/
        var dpos = 2;   /* data pos*/
        var type = -1;
        var prev = buffer[0];
        var tp = 0;
        var s;
        var v, v1;

        compressedBuff[0] = prev >> 8;
        compressedBuff[1] = prev & 0xFF;

        /* stats*/
        /*var stat = [0, 0, 0, 0, 0, 0, 0, 0];
        var med = [0, 0, 0, 0, 0, 0, 0, 0];
        var maxs = [0, 0, 0, 0, 0, 0, 0, 0];

        function calcStats()
        {
            if (type != -1)
            { 
                var c = ++stat[type];
                var sz = compressedBuff[pos] & 0x1F;
                med[type] = (med[type] * (c - 1) + sz) / c;
                if (maxs[type] < sz) maxs[type] = sz;
            }
        }*/
        
        for (var i = 1; i < buffer.length; i++)
        {
            var val = buffer[i];
            var dist = val - prev;
            prev = val;
            tmpBuff[tp++] = dist;
            num++;
            maxd |= dist < 0 ? -dist : dist;

            if (num == 1 && maxd == 0) /* zero (no signal change, compression 2 -> 62)*/
            {
                if (type != 7 || (s = compressedBuff[pos] & 0x1F) >= 31)
                {
                    /*calcStats();*/
                    pos = dpos++;
                    type = 7;
                    s = 0;
                }
                s++;
                /* set mark*/
                compressedBuff[pos] = type << 5 | s;
                /*------*/
                num = 0;
                tp = 0;
                maxd = 0;
            }
            if (num == 2)
            {   
                if (maxd < 8) /* 4 bit (compresstion 2 -> 3.875)*/
                {
                    if (type != 6 || (s = dpos - pos - 1) >= 31)
                    {
                        /*calcStats();*/
                        pos = dpos++;
                        type = 6;
                        s = 0;
                    }
                    s++;
                    /* set mark*/
                    compressedBuff[pos] = type << 5 | s;
                    /* fill data*/
                    compressedBuff[dpos++] = (tmpBuff[0] + 7) << 4 | (tmpBuff[1] + 7);
                    /*------*/
                    num = 0;
                    tp = 0;
                    maxd = 0;
                }
                else if (maxd < 128 && maxd > 31) /* 8 bit (compresstion 1.33(3) -> 1.9787234)*/
                {
                    if (type != 4 || (s = (dpos - pos - 1) / 2) >= 31)
                    {
                        /*calcStats();*/
                        pos = dpos++;
                        type = 4;
                        s = 0;
                    }
                    s++;
                    /* set mark*/
                    compressedBuff[pos] = type << 5 | s;
                    
                    /* fill data*/
                    compressedBuff[dpos++] = tmpBuff[0] + 127;
                    compressedBuff[dpos++] = tmpBuff[1] + 127;
                    /*------*/
                    num = 0;
                    tp = 0;
                    maxd = 0;
                }
            }
            else if (num == 4)
            {
                if (maxd < 32) /* 6 bit  (compresstion 2 -> 2.6382979)*/
                {
                    if (type != 5 || (s = (dpos - pos - 1) / 3) >= 31)
                    {
                        /*calcStats();*/
                        pos = dpos++;
                        type = 5;
                        s = 0;
                    }
                    s++;
                    /* set mark*/
                    compressedBuff[pos] = type << 5 | s;
                    /* fill data*/
                    v = tmpBuff[0] + 31;
                    v1 = tmpBuff[1] + 31;
                    compressedBuff[dpos++] = v << 2 | v1 >> 4;
                    v = tmpBuff[2] + 31;
                    compressedBuff[dpos++] = (v1 & 0xF) << 4 | v >> 2;
                    v1 = tmpBuff[3] + 31;
                    compressedBuff[dpos++] = (v & 0x3) << 6 | v1;
                }
                else if (maxd < 512)  /* 10 bit  (compresstion 1.33(3) -> 1.58974359)*/
                {
                    if (type != 3 || (s = (dpos - pos - 1) / 5) >= 31)
                    {
                        /*calcStats();*/
                        pos = dpos++;
                        type = 3;
                        s = 0;
                    }
                    s++;
                    /* set mark*/
                    compressedBuff[pos] = type << 5 | s;
                    /* fill data*/
                    v = tmpBuff[0] + 511; /* to uint*/
                    v1 = tmpBuff[1] + 511;
                    compressedBuff[dpos++] = v >> 2;
                    compressedBuff[dpos++] = (v & 0x03) << 6 | v1 >> 4;
                    v = tmpBuff[2] + 511;
                    compressedBuff[dpos++] = (v1 & 0x0F) << 4 | v >> 6;
                    v1 = tmpBuff[3] + 511;
                    compressedBuff[dpos++] = (v & 0x3F) << 2 | v1 >> 8;
                    compressedBuff[dpos++] = v1 & 0xFF;
                }
                else if (maxd < 2048)  /* 12 bit  (compresstion 1.14 -> 1.326203)*/
                {
                    if (type != 2 || (s = (dpos - pos - 1) / 6) >= 31)
                    {
                        /*calcStats();*/
                        pos = dpos++;
                        type = 2;
                        s = 0;
                    }
                    s++;
                    /* set mark*/
                    compressedBuff[pos] = type << 5 | s;
                    /*console.log(i,' > ', tmpBuff[0] + 2047, tmpBuff[1] + 2047, tmpBuff[2] + 2047, tmpBuff[3] + 2047);*/
                    /* fill data*/
                    v = tmpBuff[0] + 2047; /* to uint*/
                    v1 = tmpBuff[1] + 2047;
                    compressedBuff[dpos++] = v >> 4;
                    compressedBuff[dpos++] = (v & 0x0F) << 4 | v1 >> 8;
                    compressedBuff[dpos++] = v1 & 0xFF;
                    v = tmpBuff[2] + 2047;
                    v1 = tmpBuff[3] + 2047;
                    compressedBuff[dpos++] = v >> 4;
                    compressedBuff[dpos++] = (v & 0x0F) << 4 | v1 >> 8;
                    compressedBuff[dpos++] = v1 & 0xFF;
                }
                else if (maxd < 8192)  /* 14 bit  (compresstion 1 -> 1.13761469)*/
                {
                    if (type != 1 || (s = (dpos - pos - 1) / 7) >= 31)
                    {
                        /*calcStats();*/
                        pos = dpos++;
                        type = 1;
                        s = 0;
                    }
                    s++;
                    /* set mark*/
                    compressedBuff[pos] = type << 5 | s;
                    /* fill data*/
                    v = tmpBuff[0] + 8191; /* to uint*/
                    v1 = tmpBuff[1] + 8191;
                    compressedBuff[dpos++] = v >> 6;
                    compressedBuff[dpos++] = (v << 2) & 0xFC | v1 >> 12;
                    compressedBuff[dpos++] = (v1 >> 4) & 0xFF;
                    v = tmpBuff[2] + 8191;
                    compressedBuff[dpos++] = (v1 << 4) & 0xF0 | v >> 10;
                    compressedBuff[dpos++] = (v >> 2) & 0xFF;
                    v1 = tmpBuff[3] + 8191;
                    compressedBuff[dpos++] = (v << 6) & 0xC0 | v1 >> 8;
                    compressedBuff[dpos++] = v1 & 0xFF;
                }
                else  /* 16 bit (without compression 0.88(8) -> 0.9959839)*/
                {
                    if (type != 0 || (s = (dpos - pos - 1) / 8) >= 31)
                    {
                        /*calcStats();*/
                        pos = dpos++;
                        type = 0;
                        s = 0;
                    }
                    s++;
                    /* set mark*/
                    compressedBuff[pos] = type << 5 | s;
                    /* fill data*/
                    v = tmpBuff[0];
                    compressedBuff[dpos++] = v >> 8;
                    compressedBuff[dpos++] = v & 0xFF;
                    v = tmpBuff[1];
                    compressedBuff[dpos++] = v >> 8;
                    compressedBuff[dpos++] = v & 0xFF;
                    v = tmpBuff[2];
                    compressedBuff[dpos++] = v >> 8;
                    compressedBuff[dpos++] = v & 0xFF;
                    v = tmpBuff[3];
                    compressedBuff[dpos++] = v >> 8;
                    compressedBuff[dpos++] = v & 0xFF;
                }
                /*------*/
                num = 0;
                tp = 0;
                maxd = 0;
            }
        }

        /* add remaining bytes*/
        if (num)
        {
            compressedBuff[dpos++] = 0;
            i = 0;
            while (num)
            {
                v = tmpBuff[i++];
                compressedBuff[dpos++] = v >> 8;
                compressedBuff[dpos++] = v & 0xFF;
                num--;
            }
        }

        
        /*calcStats();
        for (i = 0; i < med.length; i++)
            med[i] = Math.round(med[i] * 100) / 100;
        console.log( stat, med, maxs );*/

        return compressedBuff.subarray(0, dpos);
    }
    
	/*
	 *	PUBLIC
	 */
	
	this.init = function (prefConfig)
	{
		objConfig = prefConfig;
		bufLen = prefConfig.bufferLength || 1024;
        origSampleRate = prefConfig.originalSampleRate || 48000;
        encoderSampleRate = prefConfig.sampleRate || 8000;

        var sampleSize;
        sampleFmt = prefConfig.sampleFormat;
        if (sampleFmt == 'S16')
        {
            buffer = new Int16Array(bufLen);
            compressedBuff = new Uint8Array(Math.floor(bufLen*2.04));
            bIs16Bit = true;
            sampleSize = 2;
        }
        else if (sampleFmt == 'U8')
        {
            buffer = new Uint8Array(bufLen);
            bIs16Bit = false;
            sampleSize = 1;
        }
        else
        {
            console.error('Broadcaster RAW encoder: unknow sample format ' + sampleFmt);
        }

        bitRate = encoderSampleRate * sampleSize * 8;

        bufPos = 0;
        nSrc = 0;
        nRes = 1;
        nLastRD = 0;
	};
    
    /**
     * Process source buffers
     */
	this.encode = function(buffers)
	{
        /* Resample source buffer to the specified sample rate (from origSampleRate to encoderSampleRate)*/
        var data = buffers[0];
        var len = data.length;
        var step = origSampleRate / encoderSampleRate;
        var i;
        
        function bufferFull()
        {
            /*onDataReady(buffer);*/
            var buf = compressBuffer(buffer);
            onDataReady(buf);
            bufPos = 0;

            /*totalCompressed += buf.length;*/
            /*totalData += (buffer.length * 2);*/
            /*console.log( '>>> size:', buf.length,' compress:', Math.round(100 * buf.length / (buffer.length * 2)), ' bitrate:', Math.round( bitRate * totalCompressed / totalData ));*/
        }

        if (bIs16Bit)
        {
            if (nLastRD)
            {
                buffer[bufPos++] =  32767 * (nLastVal * nLastRD + data[0] * (1 - nLastRD)); /* convert Float32 [-1 .. 1] to Int16 value*/
                nLastRD = 0;
                if (bufPos >= bufLen) bufferFull();
                nRes++;
            }

            for (i = 0; i < len; i++)
            {
                nSrc++;
                if (nSrc / nRes >= step)
                {
                    var d = (nSrc - nRes * step);
                    if (d > 0)
                    {
                        if (i + 1 < len)
                        {
                            buffer[bufPos++] = 32767 * (data[i] * d + data[i + 1]*(1 - d)); /* convert Float32 [-1 .. 1] to Int16 value*/
                        }
                        else
                        {
                            nLastVal = data[i];
                            nLastRD = d;
                            break;
                        }
                    }
                    else
                    {
                        buffer[bufPos++] = 32767 * data[i]; /* convert Float32 [-1 .. 1] to Int16 value*/
                    }

                    if (bufPos >= bufLen) bufferFull();
                    nRes++;
                }
            }
        }
        else
        {
            for (i = 0; i < len; i++)
            {
                nSrc++;
                if (nSrc / nRes >= step)
                {
                    buffer[bufPos++] = Math.round(255 * Math.max(0, Math.min(1, 0.5 + 0.5 * data[i]))); /* convert Float32 [-1 .. 1] to Uint8 value*/

                    if (bufPos >= bufLen)
                    {
                        onDataReady(buffer);
                        bufPos = 0;
                    }
                    nRes++;
                }
            }
        }
	};
	
	this.getSampleRate = function() { return encoderSampleRate; };
	
	this.getBitRate = function() { return bitRate / 1000; };
	
    this.getNumChannels = function() { return 1; };
    
    this.getBufferLength = function() { return bufLen; };

    this.getFormat = function() { 
        return 'raw:' + this.getNumChannels() 
            +':'+ sampleFmt 
            +':'+ this.getSampleRate() 
            +':'+ this.getBufferLength(); 
    }
}

/**
 *	Audio capture class.
 */
function WebAudioCapture()
{
	/* fork getUserMedia for multiple browser versions, for those*/
	/* that need prefixes*/
	var objProcessor			= null;
	var objMicrophone			= null;
	var configs					= null;
    var objGainNode				= null;
    var bStarted                = false;
    var self = this;
						  
	/**
	 *	Processes audio stream.
	 */
	function processAudioStream(mediaStream)
	{
        var objAudioContext	= configs.audioContext;

        try
        {
            objGainNode = objAudioContext.createGain();
            objGainNode.gain.value = configs.volume;
            
            objMicrophone = objAudioContext.createMediaStreamSource(mediaStream);
            objProcessor = objAudioContext.createScriptProcessor(configs.bufferLength, configs.numInputChannels, configs.numOutputChannels);
            objProcessor.onaudioprocess = function(event)
            {
                if (!bStarted) {
                    bStarted = true;
                    if (!configs.onStartCapture(event.inputBuffer.sampleRate)) {
                        self.stopAudioCapture();
                        return;
                    }
                }

                var aBuffers = [];
                for (var i = 0; i < event.inputBuffer.numberOfChannels; i++)
                    aBuffers[i] = event.inputBuffer.getChannelData(i);
                configs.onAudioData(aBuffers);
            };
            objMicrophone.connect(objGainNode);
            objGainNode.connect(objProcessor);
            objProcessor.connect(objAudioContext.destination);
        }
        catch(e)
        {
            configs.onErrorCapture('connect_media_source_error', e);
        }
	}
						  
	/**
	 *	Starts capture of an audio stream.
	 */
	this.startAudioCapture = function(conf)
	{
        configs = conf;

        var audioStreamConstraints = {
            optional: [
                {echoCancellation: false},
                {autoGainControl: false},
                {noiseSuppression: false},
                {exposureCompensation: false},
                {googAutoGainControl: false}, 
                {googAutoGainControl2: false}, 
                {googEchoCancellation: false},
                {googEchoCancellation2: false},
                {googNoiseSuppression: false},
                {googNoiseSuppression2: false},
                {googHighpassFilter: false},
                {googAudioMirroring: false},
                {mozAutoGainControl: false},
                {mozNoiseSuppression: false}]
            ,mozAutoGainControl: false
            ,mozNoiseSuppression: false
        };

        bStarted = false;

        if (navigator.mediaDevices && typeof navigator.mediaDevices.getUserMedia == 'function')
        {
            navigator.mediaDevices.getUserMedia({ audio:audioStreamConstraints })
                .then(processAudioStream)
                .catch(function(){ configs.onErrorCapture('get_media_error'); });
			return true;
        }
        else
        {
            navigator.vbcGetUserMedia = (navigator.getUserMedia ||
                navigator.webkitGetUserMedia ||
                navigator.mozGetUserMedia ||
                navigator.msGetUserMedia);
            
            if (navigator.vbcGetUserMedia)
            {
                navigator.vbcGetUserMedia({ audio:audioStreamConstraints },
                    processAudioStream, 
                    function(){ configs.onErrorCapture('get_media_error'); }
                    );
                return true;
            }
        }
		return false;
	};
	
	/**
	 *	Stops capture of an audio stream.
	 */
	this.stopAudioCapture = function()
	{
		if (objMicrophone)
		{
			console.log('Stop audio capture');
			objProcessor.disconnect();
			objMicrophone.disconnect();
			objGainNode.disconnect();
			objProcessor = null;
			objMicrophone = null;
			objEncoder = null;
            objGainNode = null;
            bStarted = false;
			configs.onStopCapture();
		}
	};
	
	/**
	 *	Sets current microphone volume.
	 *	@param volume	Volume value [0..1]. 
	 */
	this.setMicVolume = function(volume)
	{
		if (objGainNode) objGainNode.gain.value = volume;
	};
}