/**
 *	Universal player of an audio stream.
 *	Can playback an icecast/shoutcast stream and a websocket stream.
 *
 * 	@param autoStart	Start playing when possible or wait for user command, true by default.
 *	@param decoderPath	Path to opus decoder file (decoderWorker.min.js by default)
 *	@param playbackType	Type of the playback: 'link' or 'stream' ('stream' is default).
 */
function AudioStreamPlayer(autoStart, decoderPath, streamType, objRadioPlayer)
{
	var _player = this;
	
	var bAutostart			= typeof(autoStart) != 'undefined' ? autoStart : true;
	var fCurVolume			= loadValue('vplayer_volume', 1);
	var bIsMuted			= loadValue('vplayer_muted', false);
	var bIsPlaying			= false;
	var sPlaybackMode		= '';
	var sStreamLink			= '';
	var objConnection		= null; /* websocket LiveConnection*/
	var nReconnectTimeout	= 5000;
	var objAudioContext		= null;
	var objGainNode			= null;
	var objAnalyzer			= null;
	var objOpusPlayer		= null;
	var objAudio			= null;
	var objAudioSource		= null;
	var objBroadcaster		= null;

	/* audio capture variables*/
	var bAudioCapture		= false;
	var nSampleRate			= 44100;
	var nBufferLength		= 2048;
	var nInputChannels		= 2;
	var nOutputChannels		= 2;
	var objCapture			= null;
	var sCallNick			= "";
	var objEncoder			= null;
	var sEncoderPath;
	
	/*
	 *	PUBLIC
	 */

	/* Events*/
	
	this.onConnect				= function(bResult) {};
	this.onPlay					= function(bResult) {};
	this.onBuffering			= function(bResult) {};
	this.onMute					= function(bMuted) {};
	this.onVolumeChanged		= function(fVolume) {};
	this.onListenersCounter		= function(nCount) {};
	this.onDescriptionChanged	= function(objData) {};
	this.onAutostartUnallowed	= function() {};
	this.onStartVisualizer		= null; /* function(objAnalyzer)*/
	this.onStopVisualizer		= null; /* function()*/
	this.onCallAccept			= null; /* function(bAccept)*/
	this.onCaptureError			= null; /* function(sErrorCode)*/
	
	/**
	 * Saves varialbe to the local storage 
	 */
	function saveValue(name, value)
	{
		localStorage.setItem(name, JSON.stringify({"val":value}));
	}
	
	/**
	 * Loads varialbe from the local storage 
	 */
	function loadValue(name, defValue)
	{
		var val = localStorage.getItem(name);
		if (val === null) return defValue;
		try{ val = JSON.parse(val);}
		catch(e) { return val; }
		return val.val;
	}


	this.connect = function(URL)
	{
		var regexp = new RegExp("^((http|https)|(ws|wss)):\/\/");
		var match = URL.match(regexp);
		
		if (objConnection)
		{
			objConnection.disconnect();
			objConnection = null;
		}
		sPlaybackMode = '';
		
		if ( match !== null )
		{
			if ( typeof(match[2]) !== "undefined")
			{
				/* It's http(s)*/
				sPlaybackMode = "link";
				_player.onBuffering(true);
				playAudioStream(URL, bAutostart);
			}
			else if ( typeof(match[3]) !== "undefined")
			{
				/* It's ws(s)*/
				_player.onBuffering(true);
				connectSocket(URL);
			}
		}
		else
		{
			console.log("Wrong protocol !!" + URL);
		}
	};

	/**
	 *	Start or continue playback.
	 *	Calls by user.
	 */
	this.play = function()
	{
		bAutostart = true;
		if (objBroadcaster)
		{
			objBroadcaster.playStream(objBroadcaster.isHearMyself());
			return;
		}
		setPlayingState(true);

		if (sPlaybackMode == "link")
		{
			_player.onBuffering(true);
			playAudioStream(sStreamLink, true);
		}
		else if ( sPlaybackMode == "stream" )
		{
			startOpusPlayback();
		}
	};

	/**
	 *	Stop playback.
	 *	Calls by user.
	 */
	this.stop = function()
	{
		bAutostart = false;
		if (objBroadcaster)
		{
			objBroadcaster.stopStream();
			return;
		}
		setPlayingState(false);
		
		if  ( sPlaybackMode == "link" )
			stopAudioStream();
		else if ( sPlaybackMode == "stream" )
			stopOpusPlayback();
	};
	
	/**
	 *	Autostart accessor. 
	 */
	this.isAutoStart = function() { return bAutostart; };
	
	/**
	 *	Stops playback and disconnects player. 
	 */
	this.disconnect = function()
	{
		setPlayingState(false);
		if  ( sPlaybackMode == "link" )
			stopAudioStream();
		else if ( sPlaybackMode == "stream" )
			stopOpusPlayback();
			
		if (objConnection)
			objConnection.disconnect(false);
	};

	/**
	 * Set volume
	 * @param fVolume gain value [0..1]
	 */
	this.setVolume = function(fVolume)
	{
		fCurVolume = fVolume;
		saveValue('vplayer_volume', fCurVolume);
		applyVolume(fVolume);
		if (bIsMuted) _player.setMute(false);
		_player.onVolumeChanged(fVolume);
	};

	/**
	 * Set mute state
	 * @param bMute
	 */
	this.setMute = function(bMute)
	{
		bIsMuted = bMute;
		saveValue('vplayer_muted', bIsMuted);
		_player.onMute(bIsMuted);
		if (bIsMuted)
			applyVolume(0);
		else
			applyVolume(fCurVolume);
	};

	/**
	 * Get playing state
	 * @returns {boolean}
	 */
	this.isPlaying = function() { return bIsPlaying; };

	/**
	 * Get current volume
	 * @returns {number}
	 */
	this.getVolume = function() { return fCurVolume; };

	/**
	 * Get mute state
	 * @returns {boolean}
	 */
	this.isMute = function() { return bIsMuted; };

	/**
	 *	Sets broadcaster. 
	 */
	this.setBroadcaster = function(broadcaster)
	{
		objBroadcaster = broadcaster;
	};
	
	/**
	 *	Starts a call to the air.
	 * 	@param nick		User nickname.  
	 */
	this.startCall = function(nick, encoderPath)
	{
		if (!objConnection.isConnected()) return false;
		sEncoderPath = encoderPath;
		sCallNick = nick;
		startAudioCapture();
		return true;
	};
	
	/**
	 *	Starts a call to the air.
	 */
	this.stopCall = function()
	{
		if (!objConnection.isConnected()) return false;
		stopAudioCapture();
		objConnection.send(JSON.stringify({cmd:'stop_call'}));
		return true;
	};
	
	/*
	 *	PRIVATE
	 */
	
	/**
	 *	Starts capture of an audio stream.
	 */
	function startAudioCapture()
	{
		if (bAudioCapture || !objConnection.isConnected()) return;
			
		bAudioCapture = true;
		bCallAccepted = false;
		if (objCapture === null) objCapture = new WebAudioCapture();
		initAudioContext();
		if (!objCapture.startAudioCapture({
				audioContext: objAudioContext,
				bufferLength: nBufferLength,
				numInputChannels: nInputChannels,
				numOutputChannels: nOutputChannels,
				volume: 1,
				/* events*/
				onAudioData: onAudioData,
				onStartCapture: function(sampleRate){
						if (!bAudioCapture || !objConnection.isConnected()) return false;
						nSampleRate = sampleRate;
						console.log("Start audio capture");
						objConnection.send(JSON.stringify({cmd:'start_call', nick:sCallNick}));
						return true;
					},
				onStopCapture: function(){ bCaptureStarted = false; },
				onErrorCapture: function(error){ if (_player.onCaptureError) _player.onCaptureError(error); }
				}))
			{if (_player.onCaptureError) _player.onCaptureError('not_supported');}
	};
	
	/**
	 *	Stops capture of an audio stream.
	 */
	function stopAudioCapture()
	{
		if (!bAudioCapture) return;
		bAudioCapture = false;
		bCallAccepted = false;
		objCapture.stopAudioCapture();
		objEncoder = null;
	};
	
	/**
	 *	Handler of encoded data. 
	 */
	function onDataEncoded(data)
	{
		if (data.length) objConnection.send(data);
	}
	
	/**
	 *	Processes audio stream.
	 */
	function onAudioData(buffers)
	{
		if (!bCallAccepted || !bAudioCapture || !objConnection.isConnected()) return;
		if (objEncoder == null)
		{	
			objEncoder = new OpusEncoderHelper(onDataEncoded, sEncoderPath);
			console.log('Init CALL encoder');
			objEncoder.init({
				numberOfChannels: buffers.length,
				sampleRate: 24000,
				originalSampleRate: nSampleRate,
				bufferLength: nBufferLength
				});
			objConnection.send(JSON.stringify({cmd:'start_call_mic', format:'ogg', channels:objEncoder.getNumChannels(),
						sample_rate:objEncoder.getSampleRate(), bit_rate:objEncoder.getBitRate()}));
		}
		objEncoder.encode(buffers);
	}

	/**
	 * Set playing state
	 * @param bPlaying
	 */
	function setPlayingState(bPlaying)
	{
		if ( bIsPlaying !== bPlaying )
		{
			bIsPlaying = bPlaying;
			_player.onPlay(bPlaying);
		}
		if (!bPlaying) { console.log('setPlayingState false'); _player.onBuffering(false); }
	}


	/**
	 * Set gain value
	 * @param fVolume
	 */
	function applyVolume(fVolume)
	{
		if (objAudio && objAudioSource == null)
			objAudio.volume = fVolume;
		if (objGainNode)
			objGainNode.gain.value = fVolume;
	}

	/**
	 * Connect to the server
	 * @param url
	 */
	function connectSocket(url)
	{
		_player.onConnect(false);
		if (!objConnection) objConnection = new LiveConnection(onWebscoketConnect, onWebsocketMessage, onWebsocketData, nReconnectTimeout);
		objConnection.connect(url);
	}
	
	/**
	 *	Handler of connection/disconnection.
	 *	@param connected	Boolean flag.
	 */
	function onWebscoketConnect(connected)
	{
		if (connected)
		{	
			_player.onConnect(true);
			objConnection.send(JSON.stringify({cmd:'add_player', type:(streamType || 'stream')}));
		}
		else
		{
			if (sPlaybackMode == "stream")
				stopOpusPlayback();
			
			if (sPlaybackMode != "link" ) _player.onConnect(false);
		}
	}
	
	/**
	 *	Processes server commands.
	 *	@param msg	A message object, received from the server.
	 */
	function onWebsocketMessage(msg)
	{
		switch (msg.cmd)
		{
			case 'start':
				stopAudioStream();
				stopOpusPlayback();
				if ( msg.type == "link" )
				{
					sPlaybackMode = "link";
					console.log("start link: " + msg.url);
					playAudioStream(msg.url, bAutostart);
				}
				else if ( msg.type == "stream" )
				{
					sPlaybackMode = "stream";

					if ( bAutostart )
					{
						setPlayingState(true);
						startOpusPlayback();
					}
					else
					{
						setPlayingState(false);
					}
				}
				break;
				
			case 'start_call_stream':
				stopAudioStream();
				stopOpusPlayback();
				sPlaybackMode = "stream";
				setPlayingState(true);
				startOpusPlayback();
			break;

			case 'set_desc':
				if (msg.hasOwnProperty("desc") && _player.onDescriptionChanged)
					_player.onDescriptionChanged(msg.desc);
				break;

			case 'counter':
				_player.onListenersCounter(msg.val);
				break;
				
			case 'set_call_accept':
				if (_player.onCallAccept) _player.onCallAccept(msg.val);			
				break;
				
			case 'set_call_status':
				switch (msg.status)
				{
					case 'accepted':
						bCallAccepted = true;
						break;
					case 'rejected':
					case 'busy':
					case 'not_respond':
					case 'none':
						stopAudioCapture();
						break;
				}
				if (_player.onCallStatus) _player.onCallStatus(msg.status);
				break;
		}
	}
	
	/**
	 *	Processes binary data.
	 *	@param data		Bitary data, received from the server.
	 */
	function onWebsocketData(data)
	{
		/* binary data (stream)*/
		if (objOpusPlayer) objOpusPlayer.playData(data);
	}
	
	/**
	 *	Create audio context and gain node.
	 */
	function initAudioContext()
	{
		if (objAudioContext === null)
		{
			var AudioContext	= window.AudioContext || window.webkitAudioContext;
			objAudioContext		= new AudioContext();
			objGainNode			= objAudioContext.createGain();
			objGainNode.connect(objAudioContext.destination);
			objGainNode.gain.value = !bIsMuted ? fCurVolume : 0;
		}
	}

	/**
	 * Prepare to play opus audio
	 */
	function startOpusPlayback()
	{
		initAudioContext();
		if (objOpusPlayer === null)
		{
			objOpusPlayer = new StreamPlayer(objAudioContext, decoderPath);
			objOpusPlayer.onBufferEmpty = onOpusBuffering;
			objOpusPlayer.onBufferFull = onOpusBufferFull;
		}
		if (_player.onStartVisualizer)
		{
			objAnalyzer = objAudioContext.createAnalyser();
			objAnalyzer.connect(objGainNode);
			objOpusPlayer.start(objAnalyzer);
			_player.onStartVisualizer(objAnalyzer);
		}
		else
		{
			objOpusPlayer.start(objGainNode);
		}

		objConnection.send(JSON.stringify({cmd: 'play_stream'}));
		_player.onBuffering(true);
	}

	/**
	 * Stop playing opus audio
	 */
	function stopOpusPlayback()
	{
		if (objAnalyzer !== null)
		{
			if (_player.onStopVisualizer) _player.onStopVisualizer();
			objAnalyzer.disconnect();
			objAnalyzer = null;
		}
		if (objOpusPlayer) objOpusPlayer.stop();
		setPlayingState(false);
	}

	/**
	 * On Opus buffering
	 */
	function onOpusBuffering()
	{
		_player.onBuffering(true);
	}

	/**
	 * On Opus buffer full
	 */
	function onOpusBufferFull()
	{
		_player.onBuffering(false);
	}

	/**
	 * Prepare to play http(s) audio stream and start playing if needed
	 * @param url
	 * @param bStart start playing or wait for play command
	 */
	function playAudioStream(url, bStart)
	{
		if (objAudio !== null) stopAudioStream();
		
		sStreamLink = url;
		
		if (!bStart) return;
		
		objAudio = objRadioPlayer || new Audio(sStreamLink);
		if (objRadioPlayer) objRadioPlayer.src = sStreamLink;
		objAudio.addEventListener("error", onAudioError);
		objAudio.addEventListener("waiting", onAudioWaiting);
		objAudio.addEventListener("playing", onAudioPlaying);
		objAudio.addEventListener("ended", onAudioEnded);
		objAudio.addEventListener("progress", onAudioProgress);
		
		if (_player.onStartVisualizer)
		{
			objAudio.crossOrigin = "anonymous";
			initAudioContext();
			objAudioSource = objAudioContext.createMediaElementSource(objAudio);
			objAnalyzer = objAudioContext.createAnalyser();
			objAudioSource.connect(objAnalyzer);
			objAnalyzer.connect(objGainNode);
			_player.onStartVisualizer(objAnalyzer);
		}
		else
		{
			/*objAudio.volume = !bIsMuted ? fCurVolume : 0;*/
            objAudio.volume = 1;
		}
		/* play audio*/
		var res = objAudio.play();
		setPlayingState(true);
		_player.onBuffering(true);
		
		if (res instanceof Promise) {
			console.log('is promise');
		    res.catch(function(error){
		        /* Check if it is the right error*/
		        console.log(error);
		        if(error.name == "NotAllowedError")
		        {
		            _player.stop();
		            _player.onAutostartUnallowed();
		        }
		    });
		}
	}

	/**
	 * Stop playing http(s) audio
	 */
	function stopAudioStream()
	{
		if (!objAudio) return;
		
		if (objAnalyzer !== null)
		{
			if (_player.onStopVisualizer) _player.onStopVisualizer();
			objAnalyzer.disconnect();
			objAnalyzer = null;
		}
		
		objAudio.removeEventListener("error", onAudioError);
		objAudio.removeEventListener("waiting", onAudioWaiting);
		objAudio.removeEventListener("playing", onAudioPlaying);
		objAudio.removeEventListener("ended", onAudioEnded);
		objAudio.removeEventListener("progress", onAudioProgress);
		
		objAudio.pause();
		objAudio.currentTime = 0;
		objAudio.src = '';
		objAudio.load();
		objAudio = null;
		if (objAudioSource)
		{
			objAudioSource.disconnect();
			objAudioSource = null;
		}
	}
	
	/**
	 *	On http(s) audio download progress
	 */
	function onAudioProgress()
	{
		/* stream connecton detected*/
		if (!objConnection) _player.onConnect(true);
		objAudio.removeEventListener("progress", onAudioProgress);
	}
	
	/**
	 *	On http(s) audio error
	 */
	function onAudioError()
	{
		if (!objConnection) _player.onConnect(false);
		stopAudioStream();
		_player.onBuffering(true);
		setTimeout(function() { playAudioStream(sStreamLink, bAutostart); }, nReconnectTimeout);
	}

	/**
	 *	On http(s) audio waiting
	 */
	function onAudioWaiting()
	{
		_player.onBuffering(true);
	}

	/**
	 *	On http(s) audio playing
	 */
	function onAudioPlaying()
	{
		_player.onBuffering(false);
		setPlayingState(true);
	}

	/**
	 *	On http(s) audio ended
	 */
	function onAudioEnded()
	{
		if (!objConnection) _player.onConnect(false);
		stopAudioStream();
		_player.onBuffering(true);
		setTimeout(function() { playAudioStream(sStreamLink, bAutostart); }, nReconnectTimeout);
	}
}