/**
 *	Stream player class.
 *	It plays a stream of Ogg Opus data.
 */
function StreamPlayer(audioContext, decoderPath)
{
	var _player = this;
	var nBufferLength			= 16384;
	var nOutputChannels			= 2;
	var objDecoder				= null;
	var aDecodedBuffers			= null;
	var bBuffered				= false;
	var objProcessor			= null;
	var AudioContext			= window.AudioContext || window.webkitAudioContext;
	var objAudioContext			= audioContext || new AudioContext();
	var nBufferTime				= 4000; /* milliseconds*/
	var nBuffers				= 16;
	/* events*/
	this.onBufferEmpty			= null;
	this.onBufferFull			= null;
	this.onBufferOverflow		= null;

	function onDataDecoded(data)
	{
		if (aDecodedBuffers === null)
		{
			aDecodedBuffers = [];
			bBuffered = false;
		}
		if (aDecodedBuffers.length < nBuffers)
		{
			aDecodedBuffers.push(data);
		}
		else
		{
			if (_player.onBufferOverflow) _player.onBufferOverflow();
			while (aDecodedBuffers.length > nBuffers/2) aDecodedBuffers.pop();
		}
	}

	function processOutput(outputBuffer)
	{
		if (!aDecodedBuffers) return;

		if (!bBuffered && aDecodedBuffers.length >= nBuffers/2)
		{
			if (_player.onBufferFull) _player.onBufferFull();
			bBuffered = true;
		}

		if (bBuffered)
		{
			if (aDecodedBuffers.length)
			{
				var decoderBuffer = aDecodedBuffers.shift();
				for (var channel = 0; channel < outputBuffer.numberOfChannels && channel < decoderBuffer.length; channel++)
					outputBuffer.getChannelData(channel).set(decoderBuffer[channel]);				
				return;
			}
			else
			{
				if (_player.onBufferEmpty) _player.onBufferEmpty();
				bBuffered = false;
			}
		}
		/* fill the output buffer with silence*/
		for (var channel = 0; channel < outputBuffer.numberOfChannels; channel++)
		{
			var buffer = outputBuffer.getChannelData(channel);
			for (var sample = 0; sample < buffer.length; sample++) buffer[sample] = 0.0;
		}
	}

	this.start = function(destination)
	{
		_player.stop();
		this.setBufferTime(nBufferTime);
		if (_player.onBufferEmpty) _player.onBufferEmpty();
		if (objDecoder === null)
		{
			objDecoder = new OpusDecoderHelper(onDataDecoded, decoderPath);
			objDecoder.init(objAudioContext.sampleRate, nBufferLength);
		}
		objProcessor = objAudioContext.createScriptProcessor(nBufferLength, 0, nOutputChannels);
		objProcessor.onaudioprocess = function(event){ processOutput(event.outputBuffer); };
		objProcessor.connect(destination || objAudioContext.destination);
	};
	
	/**
	 *	Sets buffer time. 
	 */
	this.setBufferTime = function(value)
	{
		nBufferTime = value;
		var bufferDuration = 1000*(nBufferLength / objAudioContext.sampleRate);
		nBuffers = 2 * Math.ceil(nBufferTime/bufferDuration);
	};

	this.getBufferTime = function()	{	return nBufferTime;		};
	
	this.stop = function()
	{
		if (objDecoder)
		{
			objDecoder.stop();
			objDecoder = null;
			objProcessor.disconnect();
			objProcessor = null;
			aDecodedBuffers = null;
			bBuffered = false;
		}
	};

	this.playData = function(arrayBuffer)
	{
		if (objDecoder) objDecoder.decode(arrayBuffer);
	};
}

/**
 *	Helper class for ogg opus decorder.
 *	You must place script 'Recorderjs-0.5.0/build/decoderWorker.min.js' with current file.
 */
function OpusDecoderHelper(onDataReady, decoderPath)
{
	var sDecoderPath	= decoderPath || 'decoderWorker.min.js';
	var objDecoder		= null;
	/*
	 *	PUBLIC
	 */
	this.init = function(desiredSampleRate, bufferLength)
	{
		this.stop();
		objDecoder = new Worker(typeof(sDecoderPath) == 'string' ? sDecoderPath : window.URL.createObjectURL(sDecoderPath));
		objDecoder.postMessage({
			command:'init',
			decoderSampleRate: 48000,
			outputBufferSampleRate: desiredSampleRate,
			bufferLength: bufferLength
		});

		objDecoder.onmessage = function(event) {
			if (event.data) onDataReady(event.data);
		};
	};

	this.decode = function(arrayBuffer)
	{
		var typedArray = new Uint8Array(arrayBuffer);
		objDecoder.postMessage({
			command: 'decode',
			pages: typedArray
		}, [typedArray.buffer] );
	};

	this.stop = function()
	{
		if (objDecoder)
		{
			objDecoder.terminate();
			objDecoder = null;
		}
	};
}